/*
FILE:     vrml_util.C
*/
/*
VERSION:  8.061
*/
/*
DATE:     8/25/2004
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2004 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <assert.h>

#include "_validation.h"
#include "color_util.h"
#include "misc_util.h"
#include "ps_util.h"

#define ROTATION_ROUND_X_AXIS  1
#define ROTATION_ROUND_Y_AXIS  2
#define ROTATION_ROUND_Z_AXIS  3

#define STICK_SEGMENTS 8
#define ANGLE_PI 3.14159265358979323846

#define CHECK if (indent_flag && buflen >= max_buflen) vrml_newline()
#define PRINT buflen += fprintf

static int needs_delimiter = FALSE;
static int indent_flag = TRUE;
static int indent = 0;
static int max_buflen = 70;
static int buflen = 0;

static COLOR default_diffusecolor = {0.8, 0.8, 0.8};
static COLOR default_emissivecolor = {0.0, 0.0, 0.0};
static COLOR default_specularcolor = {0.0, 0.0, 0.0};

static int def_count = 0;

static FILE *vrml_outfile;

static void draw_ribbons(const int _n_ribbons, _RIBBONS **_ribbons,
                 const double rot[4][4], const int color_type);
static void draw_bonds(const int _n_bond, _LINK_BOND **bonds, const double rot[4][4],
                 const int color_type);
static void draw_blocks(const int _n_blocks, _BLOCK **_blocks, const double m[4][4]);
static void draw_crystal(COORD *base);
static void vrml_start_plot(const int add_rotation);
static void vrml_finist_plot(const int add_rotation);
static void output_appearance(const int is_line, const int col);
static void pad_blanks(int n);
static void put_delimiter(void);
static void vrml_add_indent(const int incr);
// static void vrml_begin_list(void);
// static void vrml_begin_node(void);
static void vrml_bbox(const COORD center, const COORD size);
static void vrml_cond_newline(void);
static void vrml_def(const char *def);
static void vrml_finish_list(void);
static void vrml_finish_node(void);
static void vrml_g(const double d);
static int  vrml_get_indent(void);
static void vrml_header(void);
static void vrml_i(const int i);
static int  vrml_invalid_def(const char *def);
static void vrml_list(const char *list);
static void vrml_material(const int dc, const int ec, const int sc,
              double ai, double sh, double tr);
static void vrml_newline (void);
static void vrml_node(const char *node);
// static void vrml_proto(char *proto);
static void vrml_qs(const char *str);
static void vrml_quad_indices(const int base, const int i1, const int i2,
                const int i3, const int i4);
static void vrml_rgb_color(const COLOR c);
static void vrml_s(const char *str);
static void vrml_s_newline(const char *str);
static void vrml_tri_indices (const int base, const int i1, const int i2, const int i3);
static void vrml_v3(const COORD v);
static void vrml_v3_g(const COORD v);
static void vrml_viewpoint (const COORD pos, /* COORD *ori, */ const double rot,
                const double fov, const char *descr);

void _VALIDATION::draw_vrml_picture(const char *filename, const double rot[4][4],
        const int color_type, COORD *base, const int add_rotation)
{
       vrml_outfile = fopen(filename, "w");
       vrml_start_plot(add_rotation);
       draw_ribbons(_n_ribbons, _ribbons, rot, color_type);
       draw_bonds(_n_link_bonds, _link_bonds, rot, color_type);
       if (_n_blocks) draw_blocks(_n_blocks, _blocks, rot);
       if (base) draw_crystal(base);
       vrml_finist_plot(add_rotation);
       fclose (vrml_outfile);
}

static void draw_ribbons(const int _n_ribbons, _RIBBONS **_ribbons,
           const double rot[4][4], const int color_type)
{
       int i, k, n_ribbon, col, base, init_col, icol, cycles, cycle;
       char string[80];
       COLOR rgb;
       _RIBBON *ribbon = NULL;

       for (k = 0; k < _n_ribbons; k++) {
            n_ribbon = _ribbons[k]->n_ribbon;
            if (n_ribbon <= 0) continue;

            ribbon = _ribbons[k]->ribbon;
            for (i = 0; i < n_ribbon; i++)
                 rotation_ribbon(ribbon[i], rot);
     
            col = ribbon[0].index;
            if (col == COLOR_BLACK || col == COLOR_GRAY) col = COLOR_WHITE;
     
            vrml_newline();
            vrml_node("Shape");
            output_appearance(FALSE, col);
            vrml_newline();
            vrml_node("geometry IndexedFaceSet");
            vrml_s_newline("creaseAngle 1");
     
            vrml_s("coord");
            sprintf(string, "MolScript_%d", def_count++);
            if (color_type == COLOR_BY_CHAIN ||
                color_type == COLOR_BY_MOLECULE) vrml_def(string);
            vrml_node("Coordinate");
            vrml_list("point");
            for (i = 0; i < n_ribbon; i++) {
                 vrml_v3(ribbon[i].p1);
                 vrml_v3(ribbon[i].p2);
                 vrml_v3(ribbon[i].p3);
                 vrml_v3(ribbon[i].p4);
            }
            vrml_finish_list();
            vrml_finish_node();
     
            vrml_newline();
            vrml_list("coordIndex");
     
            for (i = 0; i < n_ribbon - 1; i++) {
                 base = 4 * i;
                 vrml_tri_indices(base, 0, 3, 4);
                 vrml_tri_indices(base, 3, 7, 4);
            }
     
            for (i = 0; i < n_ribbon - 1; i++) {
                 base = 4 * i;
                 vrml_tri_indices(base, 2, 1, 5);
                 vrml_tri_indices(base, 5, 6, 2);
            }
       
            if (color_type == COLOR_BY_CHAIN ||
                color_type == COLOR_BY_MOLECULE) {
                 vrml_finish_list();
                 vrml_finish_node();
                 vrml_finish_node();
     
                 vrml_newline();
                 vrml_node("Shape");
                 output_appearance(FALSE, COLOR_GRAY);
                 vrml_newline();
                 vrml_node("geometry IndexedFaceSet");
                 vrml_s_newline("creaseAngle 0.79");
                 vrml_s("coord USE");
                 vrml_s_newline(string);
                 vrml_list("coordIndex");
            }
     
            vrml_quad_indices(0, 0, 1, 2, 3);
     
            for (i = 0; i < n_ribbon - 1; i++) {
                 base = 4 * i;
                 vrml_tri_indices(base, 1, 0, 5);
                 vrml_tri_indices(base, 0, 4, 5);
            }
     
            for (i = 0; i < n_ribbon - 1; i++) {
                 base = 4 * i;
                 vrml_tri_indices(base, 2, 6, 3);
                 vrml_tri_indices(base, 3, 6, 7);
            }
            base = 4 * (n_ribbon - 1);
            vrml_quad_indices(base, 0, 1, 2, 3);
     
            vrml_finish_list();
     
            if (color_type != COLOR_BY_CHAIN &&
                color_type != COLOR_BY_MOLECULE) {
                 vrml_newline();
                 vrml_s_newline("colorPerVertex FALSE");
                 vrml_node("color Color");
                 vrml_list("color");
     
                 get_rgb_color_from_index(col, rgb);
                 vrml_rgb_color(rgb);
     
                 init_col = col;
                 for (i = 1; i < n_ribbon - 1; i++) {
                      icol = ribbon[i].index;
                      if (icol == COLOR_BLACK || icol == COLOR_GRAY)
                           icol = COLOR_WHITE;
                      if (icol != init_col) {
                           init_col = icol;
                           get_rgb_color_from_index(icol, rgb);
                           vrml_rgb_color(rgb);
                      }
                 }
                 vrml_finish_list();
                 vrml_finish_node();
     
                 vrml_newline();
                 vrml_list("colorIndex");
                 cycles = 2;
                 for (cycle = 0; cycle < cycles; cycle++) {
                      init_col = col;
                      base = 0;
                      vrml_i(base);
                      vrml_i(base);
                      for (i = 1; i < n_ribbon - 1; i++) {
                           icol = ribbon[i].index;
                           if (icol == COLOR_BLACK || icol == COLOR_GRAY)
                                icol = COLOR_WHITE;
                           if (icol != init_col) {
                                init_col = icol;
                                base++;
                           }
                           vrml_i(base);
                           vrml_i(base);
                      }
                 }
                 vrml_i(0);
                 for (cycle = 0; cycle < cycles; cycle++) {
                      base = 0;
                      init_col = col;
                      for (i = 0; i < n_ribbon - 1; i++) {
                           icol = ribbon[i].index;
                           if (icol == COLOR_BLACK || icol == COLOR_GRAY)
                                icol = COLOR_WHITE;
                           if (icol != init_col) {
                                init_col = icol;
                                base++;
                           }
                           vrml_i(base);
                           vrml_i(base);
                      }
                 }
                 vrml_i(base);
                 vrml_finish_list();
            }
     
            vrml_finish_node();
            vrml_finish_node();
       }
}

static void draw_bonds(const int _n_bond, _LINK_BOND **_bonds, const double rot[4][4],
        const int color_type)
{
       int i, col, init_col, icol, base;
       COLOR rgb;
       COORD coord;

       if (_n_bond == 0 || _bonds == NULL) return;
       col = _bonds[0]->color_a;
       if (col == COLOR_BLACK || col == COLOR_GRAY) col = COLOR_WHITE;

       vrml_newline();
       vrml_node("Shape");
       output_appearance(TRUE, col);
       vrml_newline();
       vrml_node("geometry IndexedLineSet");
       vrml_s("coord");
       vrml_node("Coordinate");
       vrml_list("point");
       for (i = 0; i < _n_bond; i++) {
            coord = _bonds[i]->a->orig();
            matrix_rotate(coord, rot);
            vrml_v3(coord);
            extent_update(coord, 0);
            coord = _bonds[i]->b->orig();
            matrix_rotate(coord, rot);
            vrml_v3(coord);
            extent_update(coord, 0);
       }
       vrml_finish_list();
       vrml_finish_node();

       vrml_newline();
       vrml_list("coordIndex");
       for (i = 0; i < _n_bond; i++) {
            vrml_i(2 * i);
            vrml_i(2*i+1);
            vrml_i(-1);
       }
       vrml_finish_list();

       if (color_type == COLOR_BY_RESIDUE) {
            vrml_newline();
            vrml_s_newline("colorPerVertex FALSE");
            vrml_node("color Color");
            vrml_list("color");

            init_col = col;
            get_rgb_color_from_index(col, rgb);
            vrml_rgb_color(rgb);
            for (i = 1; i < _n_bond; i++) {
                 icol = _bonds[i]->color_a;
                 if (icol == COLOR_BLACK || icol == COLOR_GRAY)
                      icol = COLOR_WHITE;
                 if (icol != init_col) {
                      init_col = icol;
                      get_rgb_color_from_index(icol, rgb);
                      vrml_rgb_color(rgb);
                 }
            }
            vrml_finish_list();
            vrml_finish_node();

            vrml_newline();
            vrml_list("colorIndex");
            base = 0;
            init_col = col;
            for (i = 0; i < _n_bond; i++) {
                 icol = _bonds[i]->color_a;
                 if (icol == COLOR_BLACK || icol == COLOR_GRAY)
                      icol = COLOR_WHITE;
                 if (icol != init_col) {
                      init_col = icol;
                      base++;
                 }
                 vrml_i(base);
            }
            vrml_finish_list();
       }
       vrml_finish_node();
       vrml_finish_node();
}

static void draw_blocks(const int _n_blocks, _BLOCK **_blocks, const double m[4][4])
{
       int i, k, col;
       char string[80];
       COORD coord;

       for (k = 0; k < _n_blocks; k++) {
            col = _blocks[k]->index;
            if (col == COLOR_BLACK || col == COLOR_GRAY) col = COLOR_WHITE;
            vrml_newline();
            vrml_node("Shape");
            output_appearance(FALSE, col);
            vrml_newline();
            vrml_node("geometry IndexedFaceSet");
            vrml_s_newline("creaseAngle 1.0");
            vrml_s("coord");
            sprintf(string, "MolScript_%d", def_count++);
            vrml_def(string);
            vrml_node("Coordinate");
            vrml_list("point");
            for (i = 0; i < 8; i++) {
                 coord = _blocks[k]->coord[i];
                 matrix_rotate(coord, m);
                 vrml_v3(coord);
            }
            vrml_finish_list();
            vrml_finish_node();

            vrml_newline();
            vrml_list("coordIndex");
            vrml_tri_indices(0, 0, 3, 4);
            vrml_tri_indices(0, 3, 7, 4);
            vrml_tri_indices(0, 2, 1, 5);
            vrml_tri_indices(0, 5, 6, 2);
            vrml_quad_indices(0, 0, 1, 2, 3);
            vrml_quad_indices(0, 0, 1, 5, 4);
            vrml_quad_indices(0, 4, 5, 6, 7);
            vrml_quad_indices(0, 6, 7, 3, 2);
            vrml_finish_list();
            vrml_finish_node();
            vrml_finish_node();
       }
}

static void draw_crystal(COORD *base)
{
       int i;
       vrml_newline();
       vrml_node("Shape");
       output_appearance(TRUE, COLOR_WHITE);
       vrml_node("geometry IndexedLineSet");
       vrml_s("coord");

       vrml_node("Coordinate");
       vrml_list("point");
       for (i = 0; i < 8; i++) {
            vrml_v3(base[i]);
            extent_update(base[i], 0);
       }
       vrml_finish_list();
       vrml_finish_node();

       vrml_newline();
       vrml_list("coordIndex");
       vrml_i(0); vrml_i(1); vrml_i(-1);
       vrml_i(1); vrml_i(2); vrml_i(-1);
       vrml_i(2); vrml_i(3); vrml_i(-1);
       vrml_i(3); vrml_i(0); vrml_i(-1);
       vrml_i(4); vrml_i(5); vrml_i(-1);
       vrml_i(5); vrml_i(6); vrml_i(-1);
       vrml_i(6); vrml_i(7); vrml_i(-1);
       vrml_i(7); vrml_i(4); vrml_i(-1);
       vrml_i(0); vrml_i(4); vrml_i(-1);
       vrml_i(1); vrml_i(5); vrml_i(-1);
       vrml_i(2); vrml_i(6); vrml_i(-1);
       vrml_i(3); vrml_i(7); vrml_i(-1);
       vrml_finish_list();

       vrml_finish_node();
       vrml_finish_node();
}

static void vrml_start_plot(const int add_rotation)
{
       vrml_header();
/*
       vrml_proto("Ball");
       vrml_s_newline("field SFVec3f p 0 0 0");
       vrml_s_newline("field SFFloat r 1.5");
       vrml_s_newline("field SFColor dc 1 1 1");
       vrml_s_newline("field SFColor ec 0 0 0");
       vrml_s_newline("field SFColor sc 0.2 0.2 0.2");
       vrml_s_newline("field SFFloat s 0");
       vrml_s("field SFFloat t 0");
       vrml_finish_list();
       vrml_newline();
       vrml_begin_node();
       vrml_node("Transform");
       vrml_s_newline("translation IS p");
       vrml_list("children");
       vrml_node("Shape");
       vrml_node("appearance Appearance");
       vrml_node("material Material");
       vrml_s_newline("diffuseColor IS dc");
       vrml_s_newline("emissiveColor IS ec");
       vrml_s_newline("specularColor IS sc");
       vrml_s_newline("shininess IS s");
       vrml_s("transparency IS t");
       vrml_finish_node();
       vrml_finish_node();
       vrml_newline();
       vrml_node("geometry Sphere");
       vrml_s("radius IS r");
       vrml_finish_node();
       vrml_finish_node();
       vrml_finish_list();
       vrml_finish_node();
       vrml_finish_node();

       vrml_proto("Cyl");
       vrml_s_newline("field SFVec3f p 0 0 0");
       vrml_s_newline("field SFRotation rot 0 0 1 0");
       vrml_s_newline("field SFFloat h 2");
       vrml_s_newline("field SFFloat r 1");
       vrml_s_newline("field SFColor dc 1 1 1");
       vrml_s_newline("field SFColor ec 0 0 0");
       vrml_s_newline("field SFColor sc 0.2 0.2 0.2");
       vrml_s_newline("field SFFloat s 0");
       vrml_s("field SFFloat t 0");
       vrml_finish_list();
       vrml_newline();
       vrml_begin_node();
       vrml_node("Transform");
       vrml_s_newline("translation IS p");
       vrml_s_newline("rotation IS rot");
       vrml_list("children");
       vrml_node("Shape");
       vrml_node("appearance Appearance");
       vrml_node("material Material");
       vrml_s_newline("diffuseColor IS dc");
       vrml_s_newline("emissiveColor IS ec");
       vrml_s_newline("specularColor IS sc");
       vrml_s_newline("shininess IS s");
       vrml_s("transparency IS t");
       vrml_finish_node();
       vrml_finish_node();
       vrml_newline();
       vrml_node("geometry Cylinder");
       vrml_s_newline("radius IS r");
       vrml_s("height IS h");
       vrml_finish_node();
       vrml_finish_node();
       vrml_finish_list();
       vrml_finish_node();
       vrml_finish_node();

       vrml_proto("Stick");
       vrml_s_newline("field SFVec3f p 0 0 0");
       vrml_s_newline("field SFRotation rot 0 0 1 0");
       vrml_s_newline("field SFVec3f c 1 1 1");
       vrml_s_newline("field SFColor dc 1 1 1");
       vrml_s_newline("field SFColor ec 0 0 0");
       vrml_s_newline("field SFColor sc 0.2 0.2 0.2");
       vrml_s_newline("field SFFloat s 0");
       vrml_s("field SFFloat t 0");
       vrml_finish_list();
       vrml_newline();
       vrml_begin_node();
       vrml_node("Transform");
       vrml_s_newline("translation IS p");
       vrml_s_newline("rotation IS rot");
       vrml_s_newline("scale IS c");
       vrml_list("children");
       vrml_node("Shape");
       vrml_node("appearance Appearance");
       vrml_node("material Material");
       vrml_s_newline("diffuseColor IS dc");
       vrml_s_newline("emissiveColor IS ec");
       vrml_s_newline("specularColor IS sc");
       vrml_s_newline("shininess IS s");
       vrml_s("transparency IS t");
       vrml_finish_node();
       vrml_finish_node();
       vrml_newline();
       vrml_node("geometry Extrusion");
       vrml_s_newline("spine [0 -1 0, 0 1 0]");
       vrml_list("crossSection");
       vrml_g(1.0);
       vrml_g(0.0);
       for(int slot = STICK_SEGMENTS - 1; slot > 0; slot--) {
         double angle = 2.0 * ANGLE_PI *(double) slot /(double) STICK_SEGMENTS;
         vrml_g(cos(angle));
         vrml_g(sin(angle));
       }
       vrml_g(1.0);
       vrml_g(0.0);
       vrml_finish_list();
       vrml_newline();
       vrml_s_newline("beginCap FALSE");
       vrml_s_newline("endCap FALSE");
       vrml_s("creaseAngle 1.7");
       vrml_finish_node();
       vrml_finish_node();
       vrml_finish_list();
       vrml_finish_node();
       vrml_finish_node();

       vrml_proto("Label");
       vrml_s_newline("field SFVec3f p 0 0 0");
       vrml_s_newline("field SFFloat sz 1");
       vrml_s_newline("field MFString c []");
       vrml_s_newline("field SFVec3f o 0 0 0");
       vrml_s_newline("field SFColor dc 1 1 1");
       vrml_s_newline("field SFColor ec 0 0 0");
       vrml_s_newline("field SFColor sc 0.2 0.2 0.2");
       vrml_s_newline("field SFFloat s 0");
       vrml_s("field SFFloat t 0");
       vrml_finish_list();
       vrml_newline();
       vrml_begin_node();
       vrml_node("Transform");
       vrml_s_newline("translation IS p");
       vrml_list("children");
       vrml_node("Billboard");
       vrml_s_newline("axisOfRotation 0 0 0");
       vrml_list("children");
       vrml_node("Transform");
       vrml_s_newline("translation IS o");
       vrml_list("children");
       vrml_node("Shape");
       vrml_node("appearance Appearance");
       vrml_node("material Material");
       vrml_s_newline("diffuseColor IS dc");
       vrml_s_newline("emissiveColor IS ec");
       vrml_s_newline("specularColor IS sc");
       vrml_s_newline("shininess IS s");
       vrml_s("transparency IS t");
       vrml_finish_node();
       vrml_finish_node();
       vrml_newline();
       vrml_node("geometry Text");
       vrml_s_newline("string IS c");
       vrml_node("fontStyle FontStyle");
       vrml_s ("size IS sz");
       vrml_finish_node();
       vrml_finish_node();
       vrml_finish_node();
       vrml_finish_list();
       vrml_finish_node();
       vrml_finish_list();
       vrml_finish_node();
       vrml_finish_list();
       vrml_finish_node();
       vrml_finish_node();
       vrml_newline();
*/
       def_count = 0;

       vrml_newline();
       vrml_node("Collision");
       vrml_s_newline("collide FALSE");
       if (vrml_get_indent() < 0) {
            vrml_list("children");
       } else {
            vrml_s("children [");
            vrml_add_indent(2);
       }

       switch (add_rotation) {
            case ROTATION_ROUND_X_AXIS:
                 vrml_newline();
                 vrml_def("Column");
                 vrml_node("Transform");
                 vrml_s_newline("rotation 1.0 0.0 0.0 0.0");
                 break;
            case ROTATION_ROUND_Y_AXIS:
                 vrml_newline();
                 vrml_def("Column");
                 vrml_node("Transform");
                 vrml_s_newline("rotation 0.0 1.0 0.0 0.0");
                 break;
            case ROTATION_ROUND_Z_AXIS:
                 vrml_newline();
                 vrml_def("Column");
                 vrml_node("Transform");
                 vrml_s_newline("rotation 0.0 0.0 1.0 0.0");
                 break;
            default:
                 break;
       }

       switch (add_rotation) {
            case ROTATION_ROUND_X_AXIS:
            case ROTATION_ROUND_Y_AXIS:
            case ROTATION_ROUND_Z_AXIS:
                 if (vrml_get_indent() < 0) {
                      vrml_list("children");
                 } else {
                      vrml_s("children [");
                      vrml_add_indent(2);
                 }
                 break;
            default:
                 break;
       }

       extent_init();
}

static void vrml_finist_plot(const int add_rotation)
{
       vrml_finish_list();

       COORD center, size;
       get_extent(center, size);
       size.x += 2.0;
       size.y += 2.0;
       size.z += 2.0;

       switch (add_rotation) {
            case ROTATION_ROUND_X_AXIS:
            case ROTATION_ROUND_Y_AXIS:
            case ROTATION_ROUND_Z_AXIS:
                 vrml_newline();
                 vrml_s("center ");
                 vrml_v3_g(center);
                 vrml_finish_node();
                 
                 vrml_newline();
                 vrml_def("Clock");
                 vrml_node("TimeSensor");
                 vrml_s_newline("cycleInterval 16.0");
                 vrml_s("loop TRUE");
                 vrml_finish_node();

                 vrml_newline();
                 vrml_def("ColumnPath");
                 vrml_node("OrientationInterpolator");
                 vrml_list("key");
                 vrml_s("0.0, 0.50, 1.0");
                 vrml_finish_list();
                 vrml_newline();
                 vrml_list("keyValue");
                 break;
            default:
                 break;
       }

       switch (add_rotation) {
            case ROTATION_ROUND_X_AXIS:
                 vrml_s_newline("1.0 0.0 0.0 0.0,");
                 vrml_s_newline("1.0 0.0 0.0 3.14,");
                 vrml_s("1.0 0.0 0.0 6.28");
                 break;
            case ROTATION_ROUND_Y_AXIS:
                 vrml_s_newline("0.0 1.0 0.0 0.0,");
                 vrml_s_newline("0.0 1.0 0.0 3.14,");
                 vrml_s("0.0 1.0 0.0 6.28");
                 break;
            case ROTATION_ROUND_Z_AXIS:
                 vrml_s_newline("0.0 0.0 1.0 0.0,");
                 vrml_s_newline("0.0 0.0 1.0 3.14,");
                 vrml_s("0.0 0.0 1.0 6.28");
                 break;
            default:
                 break;
       }

       switch (add_rotation) {
            case ROTATION_ROUND_X_AXIS:
            case ROTATION_ROUND_Y_AXIS:
            case ROTATION_ROUND_Z_AXIS:
                 vrml_finish_list();
                 vrml_finish_node();
                 vrml_finish_list();
                 break;
            default:
                 break;
       }

       vrml_bbox(center, size);
       vrml_finish_node();

       switch (add_rotation) {
            case ROTATION_ROUND_X_AXIS:
            case ROTATION_ROUND_Y_AXIS:
            case ROTATION_ROUND_Z_AXIS:
                 vrml_newline();
                 vrml_s_newline("ROUTE Clock.fraction_changed   TO ColumnPath.set_fraction");
                 vrml_s_newline("ROUTE ColumnPath.value_changed TO Column.set_rotation");
                 break;
            default:
                 break;
       }

       vrml_newline();
       vrml_node("NavigationInfo");
       vrml_s("speed ");
       vrml_g(4.0);
       vrml_newline();
       vrml_list("type");
       vrml_qs("EXAMINE");
       vrml_qs("FLY");
       vrml_finish_list();
       vrml_finish_node();

       vrml_newline();
       if (size.x > size.y) {
            center.z += 2.0 * size.x;
       } else {
            center.z += 2.0 * size.y;
       }
       vrml_viewpoint(center, /* NULL, */ 0.0, 0.0, "overall default");
       vrml_newline();
}

static void output_appearance(const int is_line, const int col)
{
       vrml_cond_newline();
       vrml_node("appearance Appearance");
       vrml_s("material");
/*
       if(current_state->colourparts) {
         if(is_line) {
           vrml_material(NULL, NULL, &(current_state->specularcolour),
                          0.2, current_state->shininess,
                          current_state->transparency);
         } else {
           vrml_material(NULL, &(current_state->emissivecolour),
                          &(current_state->specularcolour),
                          0.2, current_state->shininess,
                          current_state->transparency);
         }
     
       } else {
         if(is_line) {
           vrml_material(NULL, c, &(current_state->specularcolour),
                          0.2, current_state->shininess,
                          current_state->transparency);
         } else {
           vrml_material(c, &(current_state->emissivecolour),
                          &(current_state->specularcolour),
                          0.2, current_state->shininess,
                          current_state->transparency);
         }
       }
*/
       if (is_line)
            vrml_material(-1, col, COLOR_BLACK, 0.2, 0.2, 0.5);
       else vrml_material(col, COLOR_BLACK, COLOR_BLACK, 0.2, 0.2, 0.5);

       vrml_finish_node();
}

static void pad_blanks(int n)
{
       assert(n >= 0);
       if (indent_flag) {
            buflen += n;
            while (n--) fputc (' ', vrml_outfile);
       }
}

static void put_delimiter(void)
{
       if (needs_delimiter) {
            fputc(' ', vrml_outfile);
            buflen++;
            needs_delimiter = FALSE;
       }
}

static void vrml_add_indent(const int incr)
{
       if (incr) {
            indent += incr;
       } else {
            indent = 0;
       }
       assert (indent >= 0);
}
/*
static void vrml_begin_list(void)
{
       put_delimiter();
       fputc('[', vrml_outfile);
       buflen++;
       vrml_add_indent(2);
       vrml_newline();
}

static void vrml_begin_node(void)
{
       put_delimiter();
       fputc ('{', vrml_outfile);
       buflen++;
       vrml_add_indent(2);
       vrml_newline();
}
*/
static void vrml_bbox(const COORD center, const COORD size)
{
       if (size.x <= 0.0) return;
       if (size.y <= 0.0) return;
       if (size.z <= 0.0) return;

       vrml_newline();
       vrml_s("bboxCenter ");
       vrml_v3_g(center);
       vrml_newline();
       vrml_s("bboxSize ");
       vrml_v3_g(size);
}

static void vrml_cond_newline(void)
{
       if (indent_flag && buflen != indent) {
            fputc ('\n', vrml_outfile);
            buflen = 0;
            needs_delimiter = FALSE;
            pad_blanks (indent);
       }
}

static void vrml_def(const char *def)
{
       // assert((int) def);
       assert(*def);
       assert(!vrml_invalid_def(def));
     
       CHECK;
       put_delimiter();
       PRINT(vrml_outfile, "DEF %s", def);
       needs_delimiter = TRUE;
}

static void vrml_finish_list(void)
{
       vrml_add_indent(-2);
       vrml_newline();
       fputc (']', vrml_outfile);
       buflen++;
       needs_delimiter = FALSE;
}

static void vrml_finish_node(void)
{
       vrml_add_indent(-2);
       vrml_newline();
       fputc('}', vrml_outfile);
       buflen++;
       needs_delimiter = FALSE;
}

static void vrml_g(const double d)
{
       CHECK;
       put_delimiter();
       PRINT (vrml_outfile, "%g", d);
       needs_delimiter = TRUE;
}

static int vrml_get_indent(void)
{
       if (indent_flag) {
            return indent;
       } else {
            return -1;
       }
}

static void vrml_header (void)
{
       needs_delimiter = FALSE;
       fprintf(vrml_outfile, "#VRML V2.0 utf8\n");
}

static void vrml_i(const int i)
{
       CHECK;
       put_delimiter();
       PRINT (vrml_outfile, "%i", i);
       needs_delimiter = TRUE;
}

static int vrml_invalid_def(const char *def)
{
       // assert((int) def);
       assert(*def);
     
       return ((strchr ("+-0123456789\"',.[]\\{}", *def) != NULL) ||
               (*(def + 1) && (strpbrk (def + 1, "\"',.[]\\{}") != NULL)));
}

static void vrml_list(const char *list)
{
       // assert((int) list);
       assert(*list);
     
       CHECK;
       put_delimiter();
       PRINT(vrml_outfile, "%s [", list);
       vrml_add_indent(2);
       vrml_newline();
}

static void vrml_material(const int dc, const int ec, const int sc,
              double ai, double sh, double tr)
{
       int prev = FALSE;

       vrml_node("Material");

       COLOR rgb;
       if (dc >= 0) {
            get_rgb_color_from_index(dc, rgb);
            if (color_unequal(rgb, default_diffusecolor)) {
                 CHECK;
                 PRINT(vrml_outfile, "diffuseColor %.2g %.2g %.2g",
                           rgb.r, rgb.g, rgb.b);
                 prev = TRUE;
            }
       }

       if (ec >= 0) {
            get_rgb_color_from_index(ec, rgb);
            if (color_unequal(rgb, default_emissivecolor)) {
                 CHECK;
                 if (prev) vrml_newline();
                 PRINT(vrml_outfile, "emissiveColor %.2g %.2g %.2g",
                           rgb.r, rgb.g, rgb.b);
                 prev = TRUE;
            }
       }

       if (sc >= 0) {
            get_rgb_color_from_index(sc, rgb);
            if (color_unequal(rgb, default_specularcolor)) {
                 CHECK;
                 if (prev) vrml_newline();
                 PRINT(vrml_outfile, "specularColor %.2g %.2g %.2g",
                           rgb.r, rgb.g, rgb.b);
                 prev = TRUE;
            }
       }

       if (fabs (ai - 0.2) <= 0.0005) {
            CHECK;
            if (prev) vrml_newline();
            PRINT (vrml_outfile, "ambientIntensity %.2g", ai);
            prev = TRUE;
       }

       if (fabs (sh - 0.2) <= 0.0005) {
            CHECK;
            if (prev) vrml_newline();
            PRINT (vrml_outfile, "shininess %.2g", sh);
            prev = TRUE;
       }

       if (tr != 0.0) {
            CHECK;
            if (prev) vrml_newline();
            PRINT (vrml_outfile, "transparency %.2g", tr);
       }

       vrml_finish_node();
}

static void vrml_newline(void)
{
       if (buflen) {
            fputc ('\n', vrml_outfile);
            buflen = 0;
       }
       needs_delimiter = FALSE;
       pad_blanks(indent);
}

static void vrml_node(const char *node)
{
       // assert((int) node);
       assert(*node);
     
       CHECK;
       put_delimiter();
       PRINT(vrml_outfile, "%s {", node);
       vrml_add_indent(2);
       vrml_newline();
}
/*
static void vrml_proto(char *proto)
{
       assert(proto);
       assert(*proto);
     
       vrml_newline();
       PRINT (vrml_outfile, "PROTO %s", proto);
       needs_delimiter = TRUE;
       vrml_begin_list();
}
*/
static void vrml_qs(const char *str)
{
       // assert ((int) str);

       CHECK;
       put_delimiter(); 
       PRINT (vrml_outfile, "\"%s\"", str);
       needs_delimiter = TRUE;
}

static void vrml_quad_indices(const int base, const int i1, const int i2,
              const int i3, const int i4)
{
       assert(base >= 0);
       assert(i1 >= 0);
       assert(i2 >= 0);
       assert(i3 >= 0);
       assert(i4 >= 0);

       CHECK;
       put_delimiter();
       PRINT(vrml_outfile, "%i %i %i %i -1", base+ i1, base+ i2,
                                             base+ i3, base+ i4);
       needs_delimiter = TRUE;
}

static void vrml_rgb_color(const COLOR c)
{
       CHECK;
       if (indent_flag) {
            if (needs_delimiter) {
                 PRINT(vrml_outfile, ", %.2g %.2g %.2g", c.r, c.g, c.b);
            } else {
                 PRINT(vrml_outfile, "%.2g %.2g %.2g", c.r, c.g, c.b);
            }
       } else {
            if (needs_delimiter) {
                 PRINT(vrml_outfile, " %.2g %.2g %.2g", c.r, c.g, c.b);
            } else {
                 PRINT(vrml_outfile, "%.2g %.2g %.2g", c.r, c.g, c.b);
            }
       }
       needs_delimiter = TRUE;
}

static void vrml_s(const char *str)
{
       // assert((int) str);
       assert(*str);
     
       CHECK;
       put_delimiter();
       PRINT(vrml_outfile, "%s", str);
       needs_delimiter = isalnum(str[strlen(str)-1]);
}

static void vrml_s_newline(const char *str)
{
       // assert((int) str);
     
       CHECK;
       put_delimiter();
       PRINT(vrml_outfile, "%s\n", str);
       buflen = 0;
       needs_delimiter = FALSE;
       pad_blanks(indent);
}

static void vrml_tri_indices (const int base, const int i1, const int i2, const int i3)
{
       assert (base >= 0);
       assert (i1 >= 0);
       assert (i2 >= 0);
       assert (i3 >= 0);
     
       CHECK;
       put_delimiter();
       PRINT (vrml_outfile, "%i %i %i -1", base + i1, base + i2, base + i3);
       needs_delimiter = TRUE;
}

static void vrml_v3(const COORD v)
{
       CHECK;
       if (indent_flag) {
            if (needs_delimiter) {
                 PRINT(vrml_outfile, ", %.2f %.2f %.2f", v.x, v.y, v.z);
            } else {
                 PRINT(vrml_outfile, "%.2f %.2f %.2f", v.x, v.y, v.z);
            }
       } else {
            if (needs_delimiter) {
                 PRINT(vrml_outfile, " %.2f %.2f %.2f", v.x, v.y, v.z);
            } else {
                 PRINT(vrml_outfile, "%.2f %.2f %.2f", v.x, v.y, v.z);
            }
       }
       needs_delimiter = TRUE;
}

static void vrml_v3_g(const COORD v)
{
       CHECK;
       if (indent_flag) {
            if (needs_delimiter) {
                 PRINT (vrml_outfile, ", %g %g %g", v.x, v.y, v.z);
            } else {
                 PRINT (vrml_outfile, "%g %g %g", v.x, v.y, v.z);
            }
       } else {
            if (needs_delimiter) {
                 PRINT (vrml_outfile, " %g %g %g", v.x, v.y, v.z);
            } else {
                 PRINT (vrml_outfile, "%g %g %g", v.x, v.y, v.z);
            }
       }
       needs_delimiter = TRUE;
}

static void vrml_viewpoint(const COORD pos, /* COORD *ori, */ const double rot,
                const double fov, const char *descr)
{
       assert (fov >= 0.0);

       vrml_newline();
       vrml_node ("Viewpoint");
       vrml_s ("position ");
       vrml_v3 (pos);
/*
       if (ori) {
            vrml_newline();
            vrml_s ("orientation ");
            vrml_v3_g (ori);
            vrml_g (rot);
       }
*/
       if (fov > 0.0) {
            vrml_newline();
            vrml_s ("fieldOfView");
            vrml_g (fov);
       }
       if (descr) {
            vrml_newline();
            vrml_s ("description ");
            vrml_qs (descr);
       }
       vrml_finish_node();
}

